// Copyright (c) 2023 Byron Lathi

/*  Async FIFO
 *
 *  see [1]
 *
 *  *_a is the input side and *_b is the output side.
 *
 * [1]  Clifford E. Cummings, “Simulation and Synthesis Techniques for Asynchronous FIFO Design,”
 *      SNUG 2002 - www.sunburst-design.com/papers/CummingsSNUG2002SJ_FIFO1.pdf
 */

module async_fifo #(
    parameter WIDTH=8,
    parameter A_SIZE=8
)(
    input i_rst_a,
    input i_clk_a,

    input i_rst_b,
    input i_clk_b,


    input logic w_en,
    input [WIDTH-1:0] i_data,
    output logic o_full,

    input r_en,
    output logic [WIDTH-1:0] o_data,
    output logic o_empty
);

localparam DEPTH = 1 << A_SIZE;

logic [A_SIZE:0] w_ptr, w_ptr_next;
logic [A_SIZE:0] w_ptr_gray, w_ptr_gray_next;
logic [A_SIZE:0] r_ptr, r_ptr_next;
logic [A_SIZE:0] r_ptr_gray, r_ptr_gray_next;

logic [A_SIZE:0] wq2_rptr_gray, rq2_wptr_gray;

logic full, empty;


logic [WIDTH-1:0] mem [DEPTH];

ff_cdc #(.WIDTH(A_SIZE+1)) rptr_ff_cdc (
    .data_a(r_ptr_gray),
    .data_b(wq2_rptr_gray),
    .clk(i_clk_a),
    .rst(i_rst_a)
);

ff_cdc #(.WIDTH(A_SIZE+1)) wptr_ff_cdc (
    .data_a(w_ptr_gray),
    .data_b(rq2_wptr_gray),
    .clk(i_clk_b),
    .rst(i_rst_b)
);

logic should_write;
assign should_write = w_en & ~o_full;

always @(posedge i_clk_a or posedge i_rst_a) begin
    if (i_rst_a) begin
        w_ptr <= '0;
        w_ptr_gray <= '0;
        o_full <= '0;
    end else begin
        w_ptr <= w_ptr_next;
        w_ptr_gray <= w_ptr_gray_next;
        o_full <= full;

        if (should_write) begin
            mem[w_ptr[A_SIZE-1:0]] <= i_data;
        end
    end
end

always_ff @(posedge i_clk_b or posedge i_rst_b) begin
    if (i_rst_b) begin
        r_ptr <= '0;
        r_ptr_gray <= '0;
        o_empty <= '0;
    end else begin
        r_ptr <= r_ptr_next;
        r_ptr_gray <= r_ptr_gray_next;
        o_empty <= empty;

        o_data <= mem[r_ptr[A_SIZE-1:0]];
    end
end

always_comb begin
    w_ptr_next = w_ptr + (w_en & ~o_full);
    w_ptr_gray_next = (w_ptr_next >> 1) ^ w_ptr_next;
    full = (w_ptr_gray_next ==
        {~wq2_rptr_gray[A_SIZE:A_SIZE-1], wq2_rptr_gray[A_SIZE-2:0]});

    r_ptr_next = r_ptr + (r_en & ~o_empty);
    r_ptr_gray_next = (r_ptr_next >> 1) ^ r_ptr_next;
    empty = (r_ptr_gray_next == rq2_wptr_gray);
end


endmodule