// simple axi ram . Only supports 32 bit aligned reads.

module axi4_lite_ram #(
    parameter RAM_INIT_FILE = "init_hex.mem",
    parameter ZERO_INIT = 0,

    parameter AXI_ADDR_WIDTH = 32,
    parameter AXI_DATA_WIDTH = 32,

    parameter [AXI_ADDR_WIDTH-1:0] BASE_ADDRESS = 0,

    parameter RAM_SIZE = 16
)(
    input i_clk,
    input i_rst,

    input i_AWVALID,
    output logic o_AWREADY,
    input [AXI_ADDR_WIDTH-1:0] i_AWADDR,
    input [2:0] i_AWPROT,

    input i_WVALID,
    output logic o_WREADY,
    input [AXI_DATA_WIDTH-1:0] i_WDATA,
    input [AXI_DATA_WIDTH/8-1:0] i_WSTRB,

    output logic o_BVALID,
    input i_BREADY,
    output logic [1:0] o_BRESP,

    input i_ARVALID,
    output logic o_ARREADY,
    input [AXI_ADDR_WIDTH-1:0] i_ARADDR,
    input [2:0] i_ARPROT,

    output logic o_RVALID,
    input i_RREADY,
    output logic [AXI_DATA_WIDTH-1:0] o_RDATA,
    output logic [1:0] o_RRESP
);

// this should depend on data width!
localparam MEMORY_DEPTH = 1 << RAM_SIZE;

logic [AXI_DATA_WIDTH-1:0] ram [MEMORY_DEPTH];


initial begin
    if (ZERO_INIT) begin
        for (int i = 0; i < MEMORY_DEPTH; i++) begin
            ram[i] = '0;
        end
    end else begin
        $readmemh(RAM_INIT_FILE, ram);
    end
end

assign o_WREADY = '1;
assign o_AWREADY = '1;

assign o_BRESP = '0;


logic [31:0] wdata;

logic [31:0] prev_data;

logic [31:0] strobe_expanded;

logic addr_valid;
logic wdata_valid;

logic [RAM_SIZE-1:0] w_addr_shift;
assign w_addr_shift = i_AWADDR[RAM_SIZE+1:2] - BASE_ADDRESS[RAM_SIZE+1:2];

logic [RAM_SIZE-1:0] r_addr_shift;
assign r_addr_shift = i_ARADDR[RAM_SIZE+1:2] - BASE_ADDRESS[RAM_SIZE+1:2];

logic [RAM_SIZE-1:0] addr;

assign o_ARREADY = i_ARVALID;

function automatic logic [31:0] strobe_expand(input logic [3:0] wstrb);
    logic [31:0] expanded;
    for (int i = 0; i < 4; i++) begin
        expanded[i*8 +: 8] = {8{wstrb[i]}};
    end

    return expanded;
endfunction

always_ff @(posedge i_clk or posedge i_rst) begin
    if (i_rst) begin
        addr_valid <= '0;
        addr <= '0;
        wdata_valid <= '0;
        wdata <= '0;
    end else begin
        o_BVALID <= i_WVALID;

        if (i_ARVALID) begin
            o_RDATA <= ram[r_addr_shift];
            o_RVALID <= '1;
            o_RRESP <= '0;
        end else begin
            o_RDATA <= '0;
            o_RVALID <= '0;
            o_RRESP <= '0;
        end

        if (i_AWVALID) begin
            addr <= w_addr_shift;
            addr_valid <= '1;
            prev_data <= ram[w_addr_shift];
        end

        if (i_WVALID) begin
            wdata <= i_WDATA;
            wdata_valid <= '1;
            strobe_expanded <= strobe_expand(i_WSTRB);
        end

        if (wdata_valid && addr_valid) begin
            ram[addr] <= (prev_data & ~strobe_expanded) | wdata & strobe_expanded;
            wdata_valid <= '0;
            addr_valid <= '0;
        end
    end
end

endmodule