// axi4_lite to apb4 converter. Supports passing through wstrb

module axi4_lite_to_apb4 #(
    parameter AXI_ADDR_WIDTH = 32,  // AXI and APB data widths must be the same
    parameter AXI_DATA_WIDTH = 32
)(
    input i_clk,
    input i_rst,

    input i_AWVALID,
    output logic o_AWREADY,
    input [AXI_ADDR_WIDTH-1:0] i_AWADDR,
    input [2:0] i_AWPROT,

    input i_WVALID,
    output logic o_WREADY,
    input [AXI_DATA_WIDTH-1:0] i_WDATA,
    input [AXI_DATA_WIDTH/8-1:0] i_WSTRB,

    output logic o_BVALID,
    input i_BREADY,
    output logic [1:0] o_BRESP,

    input i_ARVALID,
    output logic o_ARREADY,
    input [AXI_ADDR_WIDTH-1:0] i_ARADDR,
    input [2:0] i_ARPROT,

    output logic o_RVALID,
    input i_RREADY,
    output logic [AXI_DATA_WIDTH-1:0] o_RDATA,
    output logic [1:0] o_RRESP,

    output logic m_apb_psel,
    output logic m_apb_penable,
    output logic m_apb_pwrite,
    output logic [2:0] m_apb_pprot,
    output logic [AXI_ADDR_WIDTH-1:0] m_apb_paddr,
    output logic [AXI_DATA_WIDTH-1:0] m_apb_pwdata,
    output logic [AXI_DATA_WIDTH/8-1:0] m_apb_pstrb,
    input  m_apb_pready,
    input  [AXI_DATA_WIDTH-1:0] m_apb_prdata,
    input  m_apb_pslverr
);

logic [AXI_ADDR_WIDTH-1:0] addr, addr_next;
logic [AXI_DATA_WIDTH-1:0] wdata, wdata_next;
logic [AXI_DATA_WIDTH/8-1:0] wstrb, wstrb_next;
logic do_read, do_read_next;

logic awvalid, awvalid_next;
logic wvalid, wvalid_next;

logic enable, enable_next;

assign m_apb_penable = enable;


always_ff @(posedge i_clk or posedge i_rst) begin
    if (i_rst) begin
        addr <= '0;
        wdata <= '0;
        wstrb <= '0;
        do_read <= '0;
        awvalid <= '0;
        wvalid <= '0;
        enable <= '0;
    end else begin
        addr <= addr_next;
        wdata <= wdata_next;
        wstrb <= wstrb_next;
        do_read <= do_read_next;
        awvalid <= awvalid_next;
        wvalid <= wvalid_next;
        enable <= enable_next;
    end
end

always_comb begin
    m_apb_psel = '0;
    enable_next = '0;
    addr_next = addr;
    wdata_next = wdata;
    wstrb_next = wstrb;
    do_read_next = do_read;
    awvalid_next = awvalid;
    wvalid_next = wvalid;
    enable_next = enable;


    m_apb_psel = '0;
    m_apb_pwrite = '0;
    m_apb_pprot = '0;
    m_apb_paddr = '0;
    m_apb_pwdata = '0;
    m_apb_pstrb = '0;

    o_RDATA = '0;
    o_RRESP = '0;
    o_RVALID = '0;
    o_ARREADY = '1;
    o_AWREADY = ~awvalid;
    o_WREADY = '1;
    o_BRESP = '0;
    o_BVALID = '0;

    if (i_ARVALID & o_ARREADY) begin
        addr_next = i_ARADDR;
        do_read_next = '1;
    end

    if (i_AWVALID & o_AWREADY) begin
        awvalid_next = '1;;
        addr_next = i_AWADDR;
    end

    if (i_WVALID) begin
        wvalid_next = '1;
        wdata_next = i_WDATA;
        wstrb_next = i_WSTRB;
    end


    if (wvalid & awvalid) begin
        m_apb_paddr = addr;
        m_apb_pwdata = wdata;
        m_apb_pstrb = wstrb;
        m_apb_pwrite = '1;
        m_apb_psel = '1;

        enable_next = '1;

        if (enable & m_apb_pready) begin
            wvalid_next = '0;
            awvalid_next = '0;
            enable_next = '0;

            // this just assumes bready
            o_BRESP = m_apb_pslverr;
            o_BVALID = '1;
        end
    end

    if (do_read) begin
        m_apb_paddr = addr;
        m_apb_pwrite = '0;
        m_apb_psel = '1;

        enable_next = '1;

        if (enable & m_apb_pready) begin
            do_read_next = '0;
            enable_next = '0;

            o_RVALID = '1;
            o_RDATA = m_apb_prdata;
        end
    end
end


endmodule
