// Copyright (c) 2023 Byron Lathi

/*  Shallow Async FIFO
 *
 *  see [1] 5.8.2
 *
 *  *_a is the input side and *_b is the output side.
 *  This fifo does not handle back to back cycle writes correctly,
 *  you need to give one cycle between each w_en to let o_wrdy
 *  update correctly.
 *
 * [1]  Clifford E. Cummings, “Clock Domain Crossing (CDC) Design & Verification
        Techniques Using SystemVerilog”, SNUG 2008 -
        http://www.sunburst-design.com/papers/CummingsSNUG2008Boston_CDC.pdf
 */

module shallow_async_fifo #(
    parameter WIDTH=8
)(
    input i_rst_a,
    input i_clk_a,

    input i_rst_b,
    input i_clk_b,


    input w_en,
    input [WIDTH-1:0] i_data,
    output logic o_wrdy,

    input r_en,
    output [WIDTH-1:0] o_data,
    output logic o_rrdy
);

localparam DEPTH = 2;

logic w_ptr, w_ptr_next;
logic r_ptr, r_ptr_next;

logic wq2_rptr, rq2_wptr;


logic [WIDTH-1:0] mem [DEPTH];

logic [WIDTH-1:0] mem_out;

ff_cdc rptr_ff_cdc (
    .data_a(r_ptr),
    .data_b(wq2_rptr),
    .clk(i_clk_a),
    .rst(i_rst_a)
);

ff_cdc wptr_ff_cdc (
    .data_a(w_ptr),
    .data_b(rq2_wptr),
    .clk(i_clk_b),
    .rst(i_rst_b)
);

always_ff @(posedge i_clk_a or posedge i_rst_a) begin
    if (i_rst_a) begin
        w_ptr <= '0;
    end else begin
        w_ptr <= w_ptr_next;
    end

    if (w_en && o_wrdy) begin
        mem[w_ptr] <= i_data;
    end
end

always_ff @(posedge i_clk_b or posedge i_rst_b) begin
    if (i_rst_b) begin
        r_ptr <= '0;
    end else begin
        r_ptr <= r_ptr_next;
    end
end

always_comb begin
    o_wrdy = ~(w_ptr ^ wq2_rptr);
    o_rrdy = (r_ptr ^ rq2_wptr);

    w_ptr_next = (w_en & o_wrdy) ^ w_ptr;
    r_ptr_next = (r_en & o_rrdy) ^ r_ptr;
end

assign o_data = mem[r_ptr];


endmodule